!+
! 
! Module: epnoisemap_mod
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: This task makes a map of the PN electronic noise.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module epnoisemap_mod

  use cal
  use dal
  use param  
  use errorHandling
  use dss
  use caloaldefs

  use binned_att
  use cube_aux
  use eimchip2sky_aux

  use cal_aux
  use constants
  use dal_aux
  use dss_ranges_aux
  use intervals_aux	! for gtiVecType
  use save_image
  use wcs_aux
use dump

  implicit none 

  private 

  public epnoisemap

  character(255) ::&
    expMapCubeSetName='',&
    templateImageSetName='',&
!    dssSetName='',&
    noiseCubeSetName='',&
    noiseImageSetName='',&
    binnedAttSetTabName = '',&
    tempImageSetName=''
  character(20) ::&
    outputStyle='',&
    attitudeStyle = '',&
    EeVRangesStyle=''
  type(RangeT), allocatable ::&
    userEeVRanges(:)
  logical ::&
    doApprox=.false.

    real(single) :: yModelPiOver300Amp, piOver300PiScaleLen
    real(single) :: powerLawAmpRate, powerLawPower

contains 

  !---------------------------------------------------------------------
  subroutine epnoisemap
  !---------------------------------------------------------------------

    implicit none 

    type(ArrayT) :: imArray
    real(single), pointer :: expMapCube(:,:,:), templateImage(:,:)
    integer :: imageNum, numAttBins,i, xiSize, yiSize, numImages
    type(DataSetT) :: set
    real(double) ::&
      expStartTimeStamp
    integer(int16) ::&
      chipXsize, chipYsize, numEeVRanges
    integer(int8) ::&
      numCcds, numNodes
!, ccdNum, nodeNum
!, ccdNodeNum, maxCcdNodeNum, ccdnr,&
!      ccdNodeCombo, numCcdNodeCombos
!    integer(int8), allocatable :: ccd2ccdnr(:,:), ccdNodeList(:)
    real(single), allocatable ::&
      noiseMapCube(:,:,:), yFunction(:), firstYRows(:,:), noiseImage(:,:), binnedAttWeights(:,:)
    type(gtiVecType), pointer :: gtiVec(:)
    type(keyInfoType), pointer :: keyInfo(:)
    type(WcsT) :: wcs
    type(RangeT), pointer :: EeVRanges(:)
    integer(int16) :: loPis(21)
    real(single) ::&
      yModelKs(12), f13Rate(12), f14Rate(12), yModelLoPiAmps(21),&
      ccdWeights(12), amp
    real(double) :: ccdExposure(12)
    character(2) :: ccdStr
    character(256) :: binnedAttSetName = ''
    character(20) :: binnedAttTabName = ''
    type(attBinType), pointer :: binnedScAtt(:)
    type(SpacecraftAttitudeType) :: medianScAttitude, dummyInstrumAttitude
    integer(int32) :: dummyInstrumentId
    logical :: noWeightColumn
logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start:

    call readParameters

    if (outputStyle .eq. 'sky') then
      !.................................................................
      ! Read some essentials from the template image:

      call message(AppMsg, verbose, "Reading data from template image.")

      set = dataSet(templateImageSetName, READ)
      imArray = array(set, 0)
      call readWcs(set, imArray, wcs) ! in ssclib/wcs_aux.

      call getCalInfo(set, dummyInstrumentId&
        , expStartTimeStamp, medianScAttitude, dummyInstrumAttitude)
      ! in ssclib/cal_aux.

      call readArrayData(imArray, templateImage)
      ! in ssclib/dal_aux.

      call release(set)

      xiSize = size(templateImage, 1)
      yiSize = size(templateImage, 2)
      deallocate(templateImage)

!*** NOTE! the only things that the template image supplies are (i) wcs, (ii) image pixel dimensions, (iii) median attitude and (iv) exposure start time stamp. It would be nice to have some facility to input the last of these in other ways.

      allocate(noiseImage(xiSize,yiSize))
      noiseImage = 0.0
    end if

    !...................................................................

    call readCubeData(expMapCubeSetName, expMapCube, keyInfo, gtiVec)
    ! in eimchip2sky/cube_aux.

    set = dataSet(expMapCubeSetName, READ)
    expStartTimeStamp = toTimeTag(stringAttribute(set, "DATE-OBS"))
    ! in caloalutils/caloaldefs.f90

    call CAL_setState(instrument=EPN)

    call release(set)

!*********** compare expMapCube and templateImage values of expStartTimeStamp??????

    chipXSize = EPN_MAX_X	! All these are defined in caloaldefs.
    chipYSize = EPN_MAX_Y
    numCcds = EPN_CCD_LAST
    numNodes = 1

if (minval(keyInfo%ccdNum) < 1) call error('',"")
if (maxval(keyInfo%ccdNum) > numCcds) call error('',"")
if (minval(keyInfo%nodeNum) < 1) call error('',"")
if (maxval(keyInfo%nodeNum) > numNodes) call error('',"")

    numImages = size(keyInfo)
if (numImages .ne. size(gtiVec)) call error('wrongNumGtiBlocks',"")
if (numImages .ne. size(expMapCube, 1)) call error('wrongNumCubeLayers',"")
if (chipXSize .ne. size(expMapCube, 2)) call error('badXSize',"")
if (chipYSize .ne. size(expMapCube, 3)) call error('badYSize',"")

if (testFlag) print *, 'ddd'

    allocate(yFunction(chipYSize))
    allocate(noiseMapCube(numImages, chipXSize, chipYSize))
    allocate(firstYRows(numImages, 2))

    if (outputStyle .eq. 'sky') then
      !.................................................................
      ! Construct the list of attitudes:

      call message(AppMsg, verbose, "Constructing attitude samples.")
      select case(attitudeStyle)
      case('binnedset')
        call splitSetTabName(binnedAttSetTabName, binnedAttSetName, binnedAttTabName)
        ! in ssclib/dal_aux.
        call readBinnedAttitude(binnedAttSetName, binnedAttTabName, binnedScAtt, noWeightColumn)
        ! in attcalc2/binned_att.
        numAttBins = size(binnedScAtt)

      case('template')
        numAttBins = 1
        allocate(binnedScAtt(1))
        call initializeBinnedAtt(binnedScAtt)
        binnedScAtt(1)%isGood = .true.
        binnedScAtt(1)%gtiDuration = 1.0d0 ! dummy value.
        binnedScAtt(1)%att = medianScAttitude

      case default
call error('badAttitudeStyle',"")
      end select

      allocate(binnedAttWeights(numImages, numAttBins))

      !.................................................................
      ! Construct the attitude vs. image table of weights:

      call message(AppMsg, verbose, "Calculating attitude weights.")

      do imageNum = 1, numImages
        call gtiWeightBinnedAttitude(gtiVec(imageNum)%gtis, binnedScAtt)
        ! in attcalc2/binned_att.

        binnedAttWeights(imageNum,:) = binnedScAtt%gtiDuration
      end do ! imageNum

!********** normalize?????
    end if

    !...................................................................

!**** get EeVRanges

!!!    call readRanges(dssPointer, 'PI', EeVRanges)
!    call copyRanges(userEeVRanges, EeVRanges) ! in ssclib/dss_ranges_aux.
allocate(EeVRanges(size(userEeVRanges)))
EeVRanges = userEeVRanges

    numEeVRanges = size(EeVRanges)

    ! Get noise templates:
    !
    call getNoiseTemplates(loPis, yModelKs, powerLawPower, powerLawAmpRate&
      , yModelLoPiAmps, ccdWeights, f13Rate, f14Rate)

    ! Construct rate cube:

    do imageNum = 1, numImages
      ! Integrate yModelLoPiAmps (+ >300 curve if necessary) to get amp:

      amp = 0.0
      if ( EeVRanges(1)%lower%value < 300.0&
      .or. EeVRanges(1)%lower%type == UNDEFINED) then
        amp = integrateRanges(real(loPis, single), yModelLoPiAmps, EeVRanges)&
          / 5.0
        ! The division by 5 is because the PI values come in multiples
        ! of 5 eV.
        ! integrateRanges is in ssclib/dss_ranges_aux.
      end if

      if ( EeVRanges(numEeVRanges)%upper%value > 300.0&
      .or. EeVRanges(numEeVRanges)%upper%type == UNDEFINED) then
        amp = amp + integrateRanges(piOver300AmpFn, EeVRanges&
          , numSamplesArg=100) / 5.0
        ! The division by 5 is because the PI values come in multiples
        ! of 5 eV.
        ! integrateRanges is in ssclib/dss_ranges_aux.
      end if

      yFunction = 0.0
      yFunction(13) = f13Rate(imageNum)
      yFunction(14) = f14Rate(imageNum)
      yFunction(15:chipYSize) = (/(exp(-yModelKs(imageNum) * real(i))&
        , i=15, chipYSize)/)
      yFunction = yFunction * ccdWeights(imageNum) * amp

      noiseMapCube(imageNum,:,:) = spread(yFunction, 1, chipXSize)&
        * expMapCube(imageNum,:,:)

      ccdExposure(imageNum) = sum(gtiVec(imageNum)%gtis%upper&
        - gtiVec(imageNum)%gtis%lower)

      write(ccdStr, '(i2.2)') imageNum

      noiseMapCube(imageNum,:,:) = noiseMapCube(imageNum,:,:) * 0.7
!*** fudge factor
    end do

    deallocate(expMapCube)

    !...................................................................
    ! Output:

    select case(outputStyle)
    case('sky')
!      call readWcs(templateImageSetName, wcs) ! in ssclib/wcs_aux.

!      call writeCubeToSkyImage(noiseMapCube&
!        , keyInfo, expStartTimeStamp&
!        , gtiVec, templateImageSetName, templateImageSetName&
!        , noiseImageSetName&
!!        , tempImageSetName, 'counts', .true., wcs, doApprox, attitudeStyle&
!!        , binnedAttSetName)
!        , tempImageSetName, 'counts', wcs, doApprox, attitudeStyle&
!        , binnedAttSetName, binnedAttTabName)
!      ! in eimchip2sky/eimchip2sky_aux.

      call rebinChipImagesToSky(noiseMapCube, keyInfo%ccdNum, wcs, binnedScAtt%att&
        , binnedAttWeights, expStartTimeStamp, 'counts', doApprox, noiseImage)

      call saveImage(noiseImage, templateImageSetName, noiseImageSetName&
        , tempImageSetName, wcs=wcs) ! in ssclib/save_image.

      deallocate(binnedScAtt)

    case('raw')
      call writeCubeData(noiseCubeSetName, expMapCubeSetName, noiseMapCube&
        , keyInfo, gtiVec) ! in eimchip2sky/cube_aux.

    case default
call error('badOutputStyle',"")
    end select

    ! Deallocate all the gtis:
    call deallocateGtiVec(gtiVec) ! in ssclib/intervals_aux.

    deallocate(EeVRanges)
    deallocate(keyInfo)

if (testFlag) print *, 'zzz epnoisemap'

  end subroutine epnoisemap

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none 

    integer :: numUserRanges, rangeNum

    !-------------------------------------------------------------------
    ! Start:

    expMapCubeSetName = stringParameter('expcubeset')
!    dssSetName = stringParameter('dssset')

    outputStyle     = stringParameter('outputstyle')
    select case(outputStyle)
    case('sky')
      templateImageSetName = stringParameter('templateset')

      attitudeStyle = stringParameter('attstyle')
      select case(attitudeStyle)
      case('binnedset')
        binnedAttSetTabName = stringParameter('binnedatttab')

      case('template')
      case default
call error('badAttStyle',"")
      end select

      noiseImageSetName = stringParameter('noiseimageset')
    case('raw')
      noiseCubeSetName = stringParameter('noisecubeset')
    case default
call error('badOutputStyle',"")
    end select

    EeVRangesStyle = stringParameter('selexprstyle')
    select case(EeVRangesStyle)
    case('userranges')
      numUserRanges = parameterCount('evlo')
if (numUserRanges .ne. parameterCount('evhi')) call error('badNumEvHi',"")
if (numUserRanges <= 0) call error('tooFewUserEvRanges',"")
      allocate(userEeVRanges(numUserRanges))
      userEeVRanges%lower%type = INCLUSIVE
      userEeVRanges%upper%type = INCLUSIVE
      do rangeNum = 1, numUserRanges
        userEeVRanges(rangeNum)%lower%value = realParameter('evlo', rangeNum-1)
        userEeVRanges(rangeNum)%upper%value = realParameter('evhi', rangeNum-1)
      end do

    case('dss')
call error('notYetSupported',"") !** just for the present.
    case default
call error('badEvRangesStyle',"")
    end select

!    tempImageSetName          = stringParameter('tempset')
tempImageSetName = 'tempset.ds'

doApprox=.true.
!****

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine getNoiseTemplates(loPis, yModelKs, powerLawPower, powerLawAmpRate&
  , yModelLoPiAmps, ccdWeights, f13Rate, f14Rate)
  !---------------------------------------------------------------------

    implicit none 

    integer(int16), intent(out) :: loPis(21)
    real(single),   intent(out) :: yModelKs(12),&
                                   powerLawPower,&
                                   powerLawAmpRate,&
                                   yModelLoPiAmps(21),&
                                   ccdWeights(12),&
                                   f13Rate(12),&
                                   f14Rate(12)

    integer(int16) :: int16to22Cts(21), i
    real(single) ::&
      powerLawAmpCts, exposure, numActiveColumns, ampsCts(12), y13Cts(12),&
      y14Cts(12), expInt(12), sumOverCcds

    !-------------------------------------------------------------------
    ! Start:

    ! For RAWY >= 15, I am going to assume an exponential model for the drop-off in count rate with RAWY, viz:
    !
    !	phi(RAWX,RAWY,E,ccd) = W(ccd) * A(E) * exp(-k(ccd)*RAWY)	(1)
    !
    ! where phi is the flux in counts per second per eV per CCD pixel; E is energy in eV (equivalent to PI units); W(ccd) is a ccd-dependent weighting factor; A is an energy-dependent amplitude; and k is a constant. For RAWY == 13 and RAWY == 14, special fluxes
    !
    !	phi_13(RAWX,E,ccd) = W(ccd) * A(E) * f_13(ccd)
    !
    ! and
    !
    !	phi_14(RAWX,E,ccd) = W(ccd) * A(E) * f_14(ccd)
    !
    ! are defined. The task of the present subroutine is to defined values of W(ccd), A(E), k(ccd), f_13(ccd) and f_14(ccd).
    !
    ! First of all, we should normalize A(E). Noting that events are only assigned energy (PI) values which are multiples of 5, we define
    !
    !	300           _60
    !	 /            \
    !	 | dE A(E)  =  > A(5n)  =  1.					(2)
    !	 /            /_
    !	200          n = 40
    !
    ! Using an event list from obs 0206990401 PN, I selected events such that (PATTERN in [1:4])&&(PI in [200:300]). For each ccd I made a histogram of counts against RAWY. The baseline value was obtained by averaging the counts from RAW=56 to RAWY=62. Once this was subtracted from the raw counts, the result (for RAWY >= 14 at least) is expected to follow
    !
    !	                       64      300
    !	                       /        / 
    !	counts(RAWY,ccd) = T * | dRAWX  | dE phi(RAWX,RAWY,E,ccd)
    !	                       /        /
    !	                       1       200
    !
    !	                 = 64T * W(ccd) * exp(-k(ccd)*RAWY),		(3)
    !
    ! where T is the exposure time. An exponential was fitted to each histogram, resulting in the following values:

    ampsCts = (/755, 729, 728, 880, 1059, 787, 517, 414, 694, 822, 641, 511/)
    yModelKs = (/0.156, 0.162, 0.167, 0.159, 0.172, 0.179, 0.142&
      , 0.123, 0.153, 0.172, 0.173, 0.157/)

    ! The k values don't need further processing but we need to extract the W values by use of equation 3:

    exposure = 8825.0
    numActiveColumns = 64.0
    ccdWeights = ampsCts / exposure / numActiveColumns

    ! The counts (after subtraction of baseline) at RAWY = 13 and 14 are as follows:

    y13Cts = (/147.6, 127.7, 109.0, 127.0, 161.0, 106.0, 131.9&
      , 156.0, 152.6, 118.7, 108.4, 116.0/)
    y14Cts = (/110.6,  79.7,  89.0, 115.0, 123.0,  81.0, 120.9&
      ,  98.0, 119.6,  86.7,  70.4,  87.0/)

    ! Since these are given by eg
    !
    !	                     64      300
    !	                     /        / 
    !	counts(13,ccd) = T * | dRAWX  | dE phi_13(RAWX,E,ccd)
    !	                     /        /
    !	                     1       200
    !
    !	               = 64T * W(ccd) * f_13(ccd),			(4)
    !
    ! the f_13, f_14 values can be obtained as follows:

    f13Rate = y13Cts / ampsCts
    f14Rate = y14Cts / ampsCts

    ! All that remains is to estimate the A(E). To do this I took the same event list, with filtering (PATTERN in [1:4])&&(RAWY in [16:22]) and made a histogram of counts against PI value (ie, energy in eV). After subtraction of counts in (PATTERN in [1:4])&&(RAWY in [56:62]) this gives:

!    int16to22Cts = (/&
!      314- 6=308,&
!      308-16=292,&
!      322-15=307,&
!      293-22=271,&
!      262-21=241,&
!      278-16=262,&
!      246-17=229,&
!      191-19=172,&
!      154-21=133,&
!      139-10=129,&
!      108-16=92,&
!       99-18=81,&
!       74- 6=68,&
!       72-12=60,&
!       41- 8=33,&
!       38- 9=29,&
!       38- 6=32,&
!       38- 5=33,&
!       32- 5=27,&
!       29- 3=26,&
!       14- 5=9&
!    /)
    int16to22Cts = (/&
      308,&
      292,&
      307,&
      271,&
      241,&
      262,&
      229,&
      172,&
      133,&
      129,&
      92,&
      81,&
      68,&
      60,&
      33,&
      29,&
      32,&
      33,&
      27,&
      26,&
      9&
    /)

    ! The energy values corresponding to the elements of int16to22Cts are:

    loPis = (/(i, i=200,300,5)/)

    ! For PI > 300 the counts dropped in a more or less power-law fashion
    !
    !	int16to22Cts(PI) = powerLawAmpCts * (PI/300)^powerLawPower,	(5)
    !
    ! with the constants being approximately given by

    powerLawAmpCts = 20.0
    powerLawPower = -8.9

    ! According to the model defined in equation 1 these count values should be given by:
    !
    !	                64      __    22
    !	                /       \     / 
    !	counts(E) = T * | dRAWX  >    | dRAWY phi(RAWX,RAWY,E,ccd)
    !	                /       /_    /
    !	                1       ccd  16
    !
    !	                       __         22
    !	                       \          /
    !	          = 64T * A(E)  >  W(ccd) | dRAWY exp(-k(ccd)*RAWY)
    !	                       /_         /
    !	                       ccd       16
    !
    !	                       __ 
    !	                       \    W(ccd)  (  -16k(ccd)     -22k(ccd))
    !	          = 64T * A(E)  >  -------- (e^          - e^         ).  (6)
    !	                       /_   k(ccd)  (                         )
    !	                       ccd
    !
    ! The A(E)s can be easily obtained from equation 6, as follows:

    expInt = (exp(-16.0*yModelKs) - exp(-22.0*yModelKs)) / yModelKs
    sumOverCcds = sum(ccdWeights * expInt)
    yModelLoPiAmps = int16to22Cts / exposure / numActiveColumns / sumOverCcds

    ! Finally, powerLawAmpCts should be treated the same way:

    powerLawAmpRate = powerLawAmpCts / exposure / numActiveColumns / sumOverCcds

  end subroutine getNoiseTemplates

  !---------------------------------------------------------------------
  function piOver300AmpFn(piVal)
  !---------------------------------------------------------------------

    implicit none 

    real(single), intent(in) :: piVal
    real(single) :: piOver300AmpFn

    !-------------------------------------------------------------------
    ! Start:

    if (piVal < 300.0) then
      piOver300AmpFn = 0.0
    else
      piOver300AmpFn = powerLawAmpRate * ((piVal / 300.0)**powerLawPower)
    end if

  end function piOver300AmpFn

  !---------------------------------------------------------------------
  function piOver300AmpFn_old(piVal)
  !---------------------------------------------------------------------

    implicit none 

    real(single), intent(in) :: piVal
    real(single) :: piOver300AmpFn_old

    !-------------------------------------------------------------------
    ! Start:

!print *, 'in piOver300AmpFn: yModelPiOver300Amp, piOver300PiScaleLen, piVal:'&
!  , yModelPiOver300Amp, piOver300PiScaleLen, piVal

    piOver300AmpFn_old = yModelPiOver300Amp&
      * exp(-(piVal-300.0) / piOver300PiScaleLen)

  end function piOver300AmpFn_old

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none 

    !-------------------------------------------------------------------
    ! Start:

  end subroutine template

end module epnoisemap_mod

