!+
! Module: temp_ssclib
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains some routines which ought to belong in ssclib.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module temp_ssclib

  use dal
  use dss

  use dss_aux
  use wcs_aux

  implicit none 

  private 

  public saveImageLogical

contains

  !---------------------------------------------------------------------
  subroutine saveImageLogical(newImage, oldImageSetName, newImageSetName&
  , tempImageSetName, xStartArg, yStartArg, wcs)
  !---------------------------------------------------------------------
  ! This subroutine creates a new set and array, copies all keywords from
  ! the old set and old 0th array, then adjusts wcs keywords if present
  ! to conform to the start pixels xStartArg and yStartArg of the new image.
  !
  ! The output array is actually int8 rather than boolean.

!*** replace the bool routine in ssclib/save_image by this!

    implicit none

    logical,  intent(in)                 :: newImage(:,:)
    character(*),   intent(in)           :: oldImageSetName,&
                                            newImageSetName,&
                                            tempImageSetName
    integer(int16), intent(in), optional :: xStartArg,&
                                            yStartArg
    type(WcsT),     intent(in), optional :: wcs

    integer(int16) ::&
      xStart, yStart
    type(DataSetT) ::&
      newSet
    type(ArrayT) ::&
      newArray
    integer(int8), pointer ::&
      arrayDataInt8(:,:)

    !-------------------------------------------------------------------
    ! Start

    if (present(xStartArg)) then
      xStart = xStartArg
    else
      xStart = 1
    end if
    if (present(yStartArg)) then
      yStart = yStartArg
    else
      yStart = 1
    end if

    call saveImageCommon(oldImageSetName, newImageSetName, tempImageSetName&
      , xStart, yStart, size(newImage, 1), size(newImage, 2), INTEGER8&
      , newSet, newArray, wcs)
    ! Can't read or write an array of type BOOLEAN yet.

    arrayDataInt8 => int8Array2Data(newArray)
    arrayDataInt8 = 0
    where(newImage) arrayDataInt8 = 1

    call release(newSet)

  end subroutine saveImageLogical

  !---------------------------------------------------------------------
  subroutine saveImageCommon(oldImageSetName, newImageSetName&
  , tempImageSetName, xStart, yStart, xSize, ySize, newDataType&
  , newSet, newArray, wcsArg)
  !---------------------------------------------------------------------

!***** add optional newPixelSize arguments.

!*** allow for input of name or block number of the ref block??


    implicit none

    character(*),   intent(in)  :: oldImageSetName,&
                                   newImageSetName,&
                                   tempImageSetName
    integer(int16), intent(in)  :: xStart,&
                                   yStart
    integer,        intent(in)  :: xSize,&
                                   ySize
    integer(int32), intent(in)  :: newDataType
    type(DataSetT), intent(out) :: newSet ! just so it can be released
    type(ArrayT),   intent(out) :: newArray
    type(WcsT),     intent(in), optional :: wcsArg

    type(DataSetT) :: refSet
    type(ArrayT) :: refArray
!    integer :: foreignStatus
    type(WcsT) :: setWcs, wcs
    logical :: wcsPresent = .false. ! default
    integer(int32), pointer ::&
      newDimVector(:)

    !-------------------------------------------------------------------
    ! Start

    if (oldImageSetName .eq. newImageSetName) then
      refSet = clone(oldImageSetName, tempImageSetName, TEMP)

    else
      refSet = dataSet(oldImageSetName, READ)
    end if
    newSet = dataSet(newImageSetName, CREATE)
    call setAttributes(attributable(newSet), attributable(refSet))

    allocate(newDimVector(0:1))
    newDimVector = (/xSize, ySize/)
    newArray = addArray(newSet, "PRIMARY", newDataType, newDimVector)
    deallocate(newDimVector)
    refArray = array(refSet, 0, READ)
    call setAttributes(attributable(newArray), attributable(refArray))

    if (hasDss(newArray)) then
      call dssDelete(block(newArray)) ! Gets rid of DSS attributes
      ! carried over from the ref set.
    end if

    call release(refSet)

    ! Regularize and adjust the wcs:

    if (present(wcsArg)) then
      call writeWcs(newArray, wcsArg) ! in ssclib/wcs_aux
      wcsPresent = .true.
    else
      if (hasWcs(newSet)) then ! set contains the wcs keywords.

        call readWcs(newSet, setWcs)
        ! Now check to see if the array block also has a set:

        if (hasWcs(newArray)) then ! array also contains the wcs keywords.
          call warning('setAlsoHasWcs','')

        else ! array doesn't contains any wcs keywords. Copy over those from
        ! the set:
          call deleteWcs(newSet) ! in ssclib/wcs_aux
          call writeWcs(newArray, setWcs) ! in ssclib/wcs_aux

          call warning('onlySetHadWcs','')
        end if

        wcsPresent = .true.

      else ! set header doesn't contain any wcs keywords. Check the array:
        if (hasWcs(newArray)) wcsPresent = .true.
      end if
    end if

    if (wcsPresent) then
      if (xStart .ne. 1) then
        call setAttribute(newArray, 'CRPIX1', wcs%X%refPixels - (xStart - 1))
      end if
      if (yStart .ne. 1) then
        call setAttribute(newArray, 'CRPIX2', wcs%Y%refPixels - (yStart - 1))
      end if
!*** all this assumes that the pixel sizes stay the same!
    end if

  end subroutine saveImageCommon

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module temp_ssclib
